%%
%% Behavioral Programming in Erlang.
%% Guy Wiener 2010 - CC-Attribution.
%% Distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND.
%%

-module(bp).
-compile(debug_info).
-export([match/1, is/1]). % Events class functions
-export([run/4, init/0, start/0, add/1, add/2, attach/1, remove/0, remove/1, 
		 bSync/1, shutdown/0, test/0]).
-import(lists, 
		[member/2, keymap/3, keyfind/3, min/1, max/1, append/1, partition/2]).
-include("bp.hrl").

init() ->
	case whereis(bp) of 
		undefined -> ok;
		_ -> shutdown()
	end,
	register(bp, spawn(bp, run, [false, [], [], []])).
	
run(Start, Procs, Others, Syncs) ->
	process_flag(trap_exit, true),
	receive
		start -> 
			run(true, Procs, Others, Syncs);
		{add, Pid, Prio} -> 
			link(Pid),
			run(Start, [{Pid, Prio} | Procs], Others, Syncs);
		{attach, Pid} ->
			run(Start, Procs, [{Pid, 0} | Others], Syncs);
		{sync, Sync} when Start -> 
			{Procs1, Syncs1} = run_sync(Procs, [Sync | Syncs]),
			run(Start, Procs1, Others, Syncs1);
		{'EXIT', Pid, _Why} ->
			remove(Pid),
			run(Start, Procs, Others, Syncs);
		shutdown ->
			[P ! shutdown || {P, _} <- Procs ++ Others], 
			unregister(bp)
	end.

start() ->
	bp ! start.

add(Fun, Prio) when is_function(Fun) ->
	F = fun() -> try Fun() catch shutdown -> ok end end,
	Pid = spawn(F),
	add(Pid, Prio);
add(Pid, Prio) when is_pid(Pid) ->
	bp ! {add, Pid, Prio},
	Pid.
add(Prio) ->
	add(self(), Prio).

attach(Pid) when is_pid(Pid) ->
	bp ! {attach, Pid},
	Pid.

remove(Pid) ->
	bp ! {sync, #rwb{pid=Pid, remove=true}}.
remove() ->
	remove(self()).

bSync(#rwb{pid=Pid, request=R, wait=W, block=B}) ->
	P = case Pid of undefined -> self(); default -> Pid end,
	W1 = case W of [] -> R; _ -> W end,
	bp ! {sync, #rwb{pid=P, request=R, wait=W1, block=B}},
	receive
		shutdown -> throw(shutdown);
		{sync, Event} -> Event
	end.

shutdown() ->
	bp ! shutdown.

event_with_min_prio(X, []) -> X;
event_with_min_prio({Min, _}, [X = {P, _} | Rest]) when P < Min ->
	event_with_min_prio(X, Rest);
event_with_min_prio(X, [_ | Rest]) ->
	event_with_min_prio(X, Rest).
event_with_min_prio(L) ->
	event_with_min_prio({?MAX_PRIO, undef}, L).

match(F) ->
	fun(E) ->
			try
				F(E),
				true
			catch
				error:function_clause -> false 
			end
	end.

is(R) -> fun(E) -> is_record(E, R) end.

run_sync(Procs, Syncs) when length(Procs) == length(Syncs) ->
	% Get the list of syncs for processes that are not removed
	{_, Keep} = partition(fun(#rwb{remove=T}) -> T end, Syncs),
	% Build a new list of {Pid, Priotiry} for processes from Keep 
	Procs1 = [{P, R} || #rwb{pid=P} <- Keep, {Q, R} <- Procs, P == Q],
	% Build a list of {Pid, Event} from syncs
	ProcReq = [{P, E} || #rwb{pid=P, request=R} <- Keep, E <- R],
	% Build a list of {Priority, Event} from the above lists
	PrioReq = keymap(
				fun(P) -> {P, R} = keyfind(P, 1, Procs), R end,
				1, ProcReq),
	% Build a flat list of blocked events
	Blocked = append([B || #rwb{block=B} <- Keep]),
	% Build a list of {Priotiry, Event} pairs for non-blocked events
	CanDo = [X || X = {_, E} <- PrioReq, not member(E, Blocked)],
	case CanDo of 
		[] -> {Procs1, Keep};  % Nothing to do, just return the new state
		_  -> 
			% Choose the event with the minimal priority
			{_Prio, Event} = event_with_min_prio(CanDo),
			%io:format("prio ~w event ~w~n", [Prio, Event]),
			% Split syncs to Released and still Waiting 
			{Rel, Wait} = partition(
							fun(#rwb{wait = W, request=R}) -> 
									case W of
										?ALL -> true;
										Pred when is_function(W) ->
											Pred(Event);
										_ -> member(Event, W) orelse
												 member(Event, R)
									end
							end,
							Keep),
			% Broadcast the sync event to released processes
			[P ! {sync, Event} || #rwb{pid = P} <- Rel],
			{Procs1, Wait}  % Return new state
	end;
run_sync(Procs, Syncs) ->
	{Procs, Syncs}.  % Nothing to do, return the same state

%% TEST MATERIAL %%

display() -> 
	Event = bSync(#rwb{wait=[morning, evening]}),
	io:format("Good ~w~n", [Event]),
	display().

morning() ->
	lists:foreach(
	  fun(_) -> bSync(#rwb{wait=[morning], request=[morning]}) end,
	  lists:seq(1,3)).

evening() ->
	lists:foreach(
	  fun(_) -> bSync(#rwb{wait=[evening], request=[evening]}) end,
	  lists:seq(1,3)).

interleave() ->
	bSync(#rwb{wait=[morning], block=[evening]}),
	bSync(#rwb{wait=[evening], block=[morning]}),
	interleave().

test() ->
	init(),
	add(spawn(fun morning/0), 1),
	add(spawn(fun display/0), 2),
	add(spawn(fun evening/0), 3),
	add(spawn(fun interleave/0), 4),
	start().
